// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package main

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"log"
	"os"
	"os/exec"
	"runtime/debug"
	"strings"
	"syscall"
)

// This is not a completely exhaustive list of allowed characters but should be sufficient.
const allowedEmail = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.-_+@"

var version = "3.5.6"

// Checks whether input string only
// consists of allowed characters.
func isAllowed(input, allowed string) bool {
	for i := 0; i < len(input); i++ {
		if !strings.ContainsAny(input, allowed) {
			return false
		}
	}

	return true
}

// This function exits the programm with the status
// code of the invoked doveadm command.
func errorHandler(err error) {
	var exitErr *exec.ExitError

	if errors.As(err, &exitErr) {
		if status, ok := exitErr.Sys().(syscall.WaitStatus); ok {
			os.Exit(status.ExitStatus())
		}
	}

	fmt.Fprintln(os.Stderr, err.Error())
	os.Exit(-1)
}

func printBuildInfo() {
	buildInfo, ok := debug.ReadBuildInfo()
	if !ok {
		log.Fatal("Can't read BuildInfo")
	}

	fmt.Println("built from commit:") //nolint:forbidigo
	fmt.Printf("  %s\n", version)     //nolint:forbidigo

	fmt.Println("Built with:")                //nolint:forbidigo
	fmt.Printf("  %s\n", buildInfo.GoVersion) //nolint:forbidigo

	fmt.Println("Dependencies:") //nolint:forbidigo
	if len(buildInfo.Deps) > 0 {
		for _, dep := range buildInfo.Deps {
			fmt.Printf("  %s %s\n", dep.Path, dep.Version) //nolint:forbidigo
		}
	} else {
		fmt.Println("  no external dependencies") //nolint:forbidigo
	}
}

func main() {
	if len(os.Args) > 1 && os.Args[1] != "" {
		behavior := os.Args[1]

		// print pwch version and build info
		if behavior == "--version" {
			printBuildInfo()
			os.Exit(0)
		}

		// terminate imap sessions
		if behavior == "kick" && os.Args[2] != "" && isAllowed(os.Args[2], allowedEmail) {
			email := os.Args[2]
			cmd := exec.Command("/usr/bin/doveadm", "kick", email) //#nosec
			err := cmd.Run()
			if err != nil {
				errorHandler(err)
			}
			os.Exit(0)
		}

		// re-encrypt keys
		if behavior == "swap" {
			var email string
			var oldHashString string
			var newHashString string

			scanner := bufio.NewScanner(os.Stdin)

			// Read email
			if scanner.Scan() {
				email = scanner.Text()
			} else {
				log.Fatal("Error reading email:", scanner.Err())
			}

			// Read oldHashString
			if scanner.Scan() {
				oldHashString = scanner.Text()
			} else {
				log.Fatal("Error reading oldHashString:", scanner.Err())
			}

			// Read newHashString
			if scanner.Scan() {
				newHashString = scanner.Text()
			} else {
				log.Fatal("Error reading newHashString:", scanner.Err())
			}

			// prevent command injection
			if !isAllowed(email, allowedEmail) {
				os.Exit(1)
			}

			cmd := exec.Command("/usr/bin/doveadm", "mailbox", "cryptokey", "password", "-u", email, "-O", "-N") //#nosec

			var input bytes.Buffer
			input.WriteString(oldHashString + "\n" + newHashString + "\n" + newHashString + "\n")

			cmd.Stdin = &input

			err := cmd.Run()
			if err != nil {
				errorHandler(err)
			}
			os.Exit(0)
		}
	}
}
