// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package main

import (
	"log"
	"net"
	"net/http"
	"os"
	"os/signal"
	"syscall"
	"time"

	"github.com/nonce9/pwch/internal/pwch"
)

var configPath = "/etc/pwch/config.yml"

func main() {
	if len(os.Args) > 1 {
		for _, arg := range os.Args {
			if arg == "--help" {
				pwch.PrintHelp()
				os.Exit(0)
			}
		}
		for _, arg := range os.Args {
			if arg == "--version" {
				pwch.PrintBuildInfo()
				os.Exit(0)
			}
		}
		if os.Args[1] == "--config" {
			configPath = os.Args[2]
		}
	}

	err := pwch.ReadConfig(&pwch.Cfg, configPath)
	if err != nil {
		log.Fatal(err)
	}

	// used for rate limiting
	pwch.LastEmailSent = time.Now()

	mux := http.NewServeMux()

	mux.HandleFunc(pwch.Cfg.URLPrefix+"/", pwch.SubmitEmailHandler)
	mux.HandleFunc(pwch.Cfg.URLPrefix+"/emailSend", pwch.EmailSendHandler)
	mux.HandleFunc(pwch.Cfg.URLPrefix+"/changePassword", pwch.PasswordChangeHandler)
	mux.HandleFunc(pwch.Cfg.URLPrefix+"/submitPassword", pwch.PasswordSubmitHandler)

	mux.Handle("/static/", http.StripPrefix("/static/", http.FileServer(http.Dir(pwch.Cfg.AssetsPath+"/css"))))

	socket, err := net.Listen("unix", pwch.Cfg.Server.SocketPath)
	if err != nil {
		log.Fatal(err)
	}

	// Cleanup the sockfile.
	c := make(chan os.Signal, 1)
	signal.Notify(c, os.Interrupt, syscall.SIGTERM)
	go func() {
		<-c
		if socketErr := os.Remove(pwch.Cfg.Server.SocketPath); socketErr != nil {
			log.Fatal(socketErr)
		}
		os.Exit(0)
	}()

	server := http.Server{ //nolint:exhaustruct
		Handler:      mux,
		ReadTimeout:  5 * time.Second,
		WriteTimeout: 10 * time.Second,
	}

	go func() {
		log.Fatal(server.Serve(socket))
	}()

	perm := os.FileMode(0o777)

	err = os.Chmod(pwch.Cfg.Server.SocketPath, perm)
	if err != nil {
		log.Fatal("Error changing socket path permissions:", err)
	}

	log.Printf("pwch %s", pwch.Version)
	log.Print("INFO: Listening on " + pwch.Cfg.Server.SocketPath)

	// delete expired one time links every 30 seconds
	ticker := time.NewTicker(30 * time.Second)
	for {
		<-ticker.C
		for k, v := range pwch.OneTimeURLs.M {
			if time.Since(v) > pwch.Cfg.OTL.ValidFor {
				pwch.DeleteFromHashMap(k)
				log.Print("INFO: Deleted expired route " + k + " from map")
			}
		}
	}
}
