// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"fmt"
	"html/template"
	"log"
	"net/http"
	"time"
)

// LastEmailSent is used for rate limiting.
var LastEmailSent time.Time

// data object for html template.
type changePasswordTemplateData struct {
	URLPrefix string
	Token     string
	Username  string
	Domain    string
	Length    int
	Lower     bool
	Upper     bool
	Digit     bool
	Special   bool
}

type submitEmailTemplateData struct {
	URLPrefix string
}

// this function will take an error message
// and execute a html template to show the user
// why his password cannot be changed.
func templatePasswordErrorPage(w http.ResponseWriter, errorMessage string) {
	tmpl, err := template.ParseFiles(Cfg.AssetsPath + "/error.html")
	if err != nil {
		log.Print(err)
	}
	err = tmpl.Execute(w, errorMessage)
	if err != nil {
		log.Print("ERROR: cannot template error page")
		log.Print(err)
	}
}

func SubmitEmailHandler(w http.ResponseWriter, _ *http.Request) {
	data := submitEmailTemplateData{
		URLPrefix: Cfg.URLPrefix,
	}

	tmpl, err := template.ParseFiles(Cfg.AssetsPath + "/submitEmail.html")
	if err != nil {
		log.Print(err)

		return
	}

	if err := tmpl.Execute(w, data); err != nil {
		log.Print(err)
		log.Print("ERROR: cannot execute template")
	}
}

func EmailSendHandler(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodPost {
		http.Error(w, "Method not allowed", http.StatusMethodNotAllowed)

		return
	}

	email := r.FormValue("email")
	if !isValidEmail(email) {
		templatePasswordErrorPage(w, "Please enter a valid email address")

		return
	}

	// rate limiting
	if time.Since(LastEmailSent) < 5*time.Second {
		http.Error(w, "Too early. Please try again.", http.StatusTooEarly)

		return
	}

	http.ServeFile(w, r, Cfg.AssetsPath+"/emailSent.html")

	enabled, mailUser, err := emailEnabled(email)
	if err != nil {
		log.Print(err)
	}

	if enabled {
		LastEmailSent = time.Now()
		go sendOneTimeLink(mailUser.Username, mailUser.Domain)
	}
}

func PasswordChangeHandler(w http.ResponseWriter, r *http.Request) {
	token := r.URL.Query().Get("token")
	username := r.URL.Query().Get("username")
	domain := r.URL.Query().Get("domain")

	url := fmt.Sprintf("changePassword?token=%s&username=%s&domain=%s", token, username, domain)

	if _, ok := OneTimeURLs.M[url]; !ok {
		fmt.Fprint(w, "Link expired")

		return
	}

	data := changePasswordTemplateData{
		URLPrefix: Cfg.URLPrefix,
		Token:     token,
		Username:  username,
		Domain:    domain,
		Length:    Cfg.PasswordPolicy.MinLength,
		Lower:     Cfg.PasswordPolicy.LowerCase,
		Upper:     Cfg.PasswordPolicy.UpperCase,
		Digit:     Cfg.PasswordPolicy.Digits,
		Special:   Cfg.PasswordPolicy.SepcialChar,
	}

	tmpl, err := template.ParseFiles(Cfg.AssetsPath + "/changePassword.html")
	if err != nil {
		log.Print(err)

		return
	}

	if err := tmpl.Execute(w, data); err != nil {
		log.Print("ERROR: cannot execute template")
		log.Print(err)
	}
}

func PasswordSubmitHandler(w http.ResponseWriter, r *http.Request) {
	token := r.URL.Query().Get("token")
	username := r.URL.Query().Get("username")
	domain := r.URL.Query().Get("domain")

	oldPass := r.FormValue("current-password")
	newPass := r.FormValue("new-password")
	confirmPass := r.FormValue("confirm-password")

	url := fmt.Sprintf("changePassword?token=%s&username=%s&domain=%s", token, username, domain)

	if _, ok := OneTimeURLs.M[url]; !ok {
		http.Redirect(w, r, "/", http.StatusFound)

		return
	}

	if err := validatePasswordFields(newPass, confirmPass, oldPass); err != nil {
		templatePasswordErrorPage(w, err.Error())

		return
	}

	if enforced, errMessage := enforcePasswordPolicy(newPass); !enforced {
		templatePasswordErrorPage(w, errMessage)

		return
	}

	if err := updatePassword(username, domain, newPass, oldPass); err != nil { //nolint:contextcheck
		templatePasswordErrorPage(w, err.Error())

		return
	}

	DeleteFromHashMap(url)
	log.Print("INFO: Deleted " + url + " from map")
	http.ServeFile(w, r, Cfg.AssetsPath+"/success.html")
}
