// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"bytes"
	"io"
	"log"
	"net/http"
	"net/http/httptest"
	"net/url"
	"os"
	"regexp"
	"strings"
	"testing"
	"time"
)

const (
	assetsPath = "../../assets/html"
	tokenURL   = "changePassword?token="
	parameters = "&username=pwch1&domain=localdomain"
)

func Test_SubmitEmailHandler(t *testing.T) {
	Cfg.AssetsPath = assetsPath

	req, err := http.NewRequest(http.MethodGet, "/", http.NoBody) //nolint:noctx
	if err != nil {
		t.Fatal(err)
	}

	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(SubmitEmailHandler)

	handler.ServeHTTP(rr, req)

	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v, want %v", status, http.StatusOK)
	}

	expected := "<title>Password Reset</title>"
	if !strings.Contains(rr.Body.String(), expected) {
		t.Errorf("handler returned unexpected body: %v not found", expected)
	}
}

func Test_EmailSendHandler(t *testing.T) {
	Cfg.DB.Host = "/run/postgresql" //nolint:goconst
	Cfg.DB.DBName = "vmail"         //nolint:goconst
	Cfg.DB.User = "vmail"
	Cfg.DB.Password = "password" //nolint:goconst
	Cfg.DB.SSLMode = "disable"

	Cfg.SMTP.Host = "localhost"
	Cfg.SMTP.Port = "587"
	Cfg.SMTP.LoginUser = "noreply@localdomain" //nolint:goconst
	Cfg.SMTP.LoginPassword = "password"
	Cfg.SMTP.Sender = "noreply@localdomain"

	Cfg.AssetsPath = assetsPath

	form := url.Values{}

	checkEmailAddress := func(_ testing.TB, expectedBody, method string, expectedCode int, pause bool) string {
		t.Helper()

		// Redirect log output to a buffer
		var buf bytes.Buffer
		log.SetOutput(&buf)

		req, err := http.NewRequest(method, "/emailSend", http.NoBody) //nolint:noctx
		if err != nil {
			t.Fatal(err)
		}

		if method == "POST" {
			req.PostForm = form
		}

		rr := httptest.NewRecorder()

		EmailSendHandler(rr, req)

		if pause {
			time.Sleep(500 * time.Millisecond)
		}
		output := buf.String()

		// Remove the date and timestamp portion from the log messages
		re := regexp.MustCompile(`\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} `)
		cleanedOutput := re.ReplaceAllString(output, "")

		// Check the response status code
		if rr.Code != expectedCode {
			t.Errorf("Expected status code %d, but got %d", expectedCode, rr.Code)
		}

		// Check the response body
		if !strings.Contains(rr.Body.String(), expectedBody) {
			t.Errorf("handler returned unexpected body: %v not found", expectedBody)
		}

		// reset log output to stdout
		log.SetOutput(os.Stdout)

		return cleanedOutput
	}

	// Test case 1
	t.Run("test valid email address", func(t *testing.T) {
		form.Add("email", "pwch1@localdomain")

		logOutput := checkEmailAddress(t, "an email may have been sent", "POST", http.StatusOK, true)

		expected := "INFO: pwch1@localdomain successfully validated\n" +
			"INFO: Sent OTL to pwch1@localdomain\n"

		if logOutput != expected {
			t.Errorf("Unexpected log output.\nExpected: %s\nActual: %s", expected, logOutput)
		}
	})

	// Test case 2
	// reset rate limiting
	LastEmailSent = time.Now().Add(-10 * time.Minute)
	form = url.Values{}
	t.Run("test invalid email address", func(t *testing.T) {
		form.Add("email", "invalid@localdomain")

		logOutput := checkEmailAddress(t, "an email may have been sent", "POST", http.StatusOK, false)

		expected := "INFO: Unknown email address: invalid@localdomain\n"

		if logOutput != expected {
			t.Errorf("Unexpected log output.\nExpected: %s\nActual: %s", expected, logOutput)
		}
	})

	// Test case 3
	LastEmailSent = time.Now()
	form = url.Values{}
	t.Run("test rate limiting", func(t *testing.T) {
		form.Add("email", "pwch1@localdomain")

		_ = checkEmailAddress(t, "Too early", "POST", http.StatusTooEarly, false)
	})

	// Test case 4
	t.Run("test wrong method", func(t *testing.T) {
		_ = checkEmailAddress(t, "Method not allowed", "GET", http.StatusMethodNotAllowed, false)
	})

	form = url.Values{}
	// Test case 5
	t.Run("test wrong email syntax", func(t *testing.T) {
		form.Add("email", "pwch1localdomain")

		_ = checkEmailAddress(t, "Please enter a valid email address", "POST", http.StatusOK, false)
	})
}

func Test_PasswordChangeHandler(t *testing.T) {
	Cfg.AssetsPath = assetsPath

	getResetPage := func(_ testing.TB, url, expected string) {
		t.Helper()

		req, err := http.NewRequest(http.MethodGet, url, http.NoBody) //nolint:noctx
		if err != nil {
			t.Fatal(err)
		}

		rr := httptest.NewRecorder()

		// Call the handler function
		PasswordChangeHandler(rr, req)

		// Check the response status code
		if rr.Code != http.StatusOK {
			t.Errorf("expected status code %d, got %d", http.StatusOK, rr.Code)
		}

		// Check the response body
		expectedBody := expected
		if !strings.Contains(rr.Body.String(), expectedBody) {
			t.Errorf("handler returned unexpected body: %v not found", expectedBody)
		}
	}

	// Test case 1
	t.Run("test with valid URL", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		getResetPage(t, testURL, "<title>Password Reset</title>")
	})

	// Test case 2
	t.Run("test with invalid URL", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters

		getResetPage(t, testURL, "Link expired")
	})
}

func Test_PasswordSubmitHandler(t *testing.T) {
	// discard log output for this function
	log.SetOutput(io.Discard)

	// set password policy
	Cfg.PasswordPolicy.MinLength = 6
	Cfg.PasswordPolicy.MaxLength = 24
	Cfg.PasswordPolicy.LowerCase = true
	Cfg.PasswordPolicy.UpperCase = true
	Cfg.PasswordPolicy.Digits = false
	Cfg.PasswordPolicy.SepcialChar = false

	Cfg.AssetsPath = assetsPath

	form := url.Values{}

	getResultPage := func(_ testing.TB, url, expectedBody string, expectedCode int) {
		t.Helper()

		req, err := http.NewRequest(http.MethodPost, "/submitPassword/"+url, http.NoBody) //nolint:noctx
		if err != nil {
			t.Fatal(err)
		}
		req.PostForm = form

		rr := httptest.NewRecorder()

		// Call the handler function
		PasswordSubmitHandler(rr, req)

		// Check the response status code
		if rr.Code != expectedCode {
			t.Errorf("expected status code %d, got %d", expectedCode, rr.Code)
		}

		// Check the response body
		if !strings.Contains(rr.Body.String(), expectedBody) {
			t.Errorf("handler returned unexpected body: %v not found", expectedBody)
		}
	}

	// Test case 1
	t.Run("test full workflow", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "password")
		form.Add("new-password", "StrongPassword123!")
		form.Add("confirm-password", "StrongPassword123!")

		getResultPage(t, testURL, "Success", http.StatusOK)
	})

	// Test case 2
	t.Run("test workflow again with same credentials and fail", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "password")
		form.Add("new-password", "StrongPassword123!")
		form.Add("confirm-password", "StrongPassword123!")

		getResultPage(t, testURL, "Current Password does not match", http.StatusOK)
	})

	// Test case 3
	form = url.Values{}
	t.Run("test redirect for expired link", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters

		form.Add("current-password", "StrongPassword1234!")
		form.Add("new-password", "StrongPassword123!+")
		form.Add("confirm-password", "StrongPassword123!+")

		getResultPage(t, testURL, "", http.StatusFound)
	})

	// Test case 4
	form = url.Values{}
	t.Run("test missmatching passwords", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "password")
		form.Add("new-password", "StrongPassword1234!")
		form.Add("confirm-password", "StrongPassword123!")

		getResultPage(t, testURL, "Passwords do not match", http.StatusOK)
	})

	// Test case 5
	form = url.Values{}
	t.Run("test setting the same password", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "password")
		form.Add("new-password", "password")
		form.Add("confirm-password", "password")

		getResultPage(t, testURL, "You are trying to set the same password again", http.StatusOK)
	})

	// Test case 6
	form = url.Values{}
	t.Run("test password policy violation", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "password")
		form.Add("new-password", "password123")
		form.Add("confirm-password", "password123")

		getResultPage(t, testURL, "Please enter at least one upper case character", http.StatusOK)
	})

	// Test case 7
	form = url.Values{}
	t.Run("revert test case 1", func(t *testing.T) {
		token, _ := genRandomString(64)
		testURL := tokenURL + token + parameters
		addToHashMap(testURL, time.Now())

		form.Add("current-password", "StrongPassword123!")
		form.Add("new-password", "password")
		form.Add("confirm-password", "password")

		Cfg.PasswordPolicy.UpperCase = false

		getResultPage(t, testURL, "Success", http.StatusOK)
	})

	// restore log output to stdout
	log.SetOutput(os.Stdout)
}
