// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"errors"
	"fmt"
	"log"
	"os"
	"runtime/debug"

	"gopkg.in/yaml.v3"
)

var Version = "3.5.6"

func PrintBuildInfo() {
	buildInfo, ok := debug.ReadBuildInfo()
	if !ok {
		log.Fatal("Can't read BuildInfo")
	}

	fmt.Println("Version:")       //nolint:forbidigo
	fmt.Printf("  %s\n", Version) //nolint:forbidigo

	fmt.Println("Built with:")                //nolint:forbidigo
	fmt.Printf("  %s\n", buildInfo.GoVersion) //nolint:forbidigo
}

func PrintHelp() {
	fmt.Println("Possible arguments:")                                                  //nolint:forbidigo
	fmt.Println("  --config		Changes default path from where to read the config file.") //nolint:forbidigo
	fmt.Println("  --help		Print this help statement.")                                 //nolint:forbidigo
	fmt.Println("  --version		Print version and build info.")                           //nolint:forbidigo
}

func ReadConfig(cfg *Config, configPath string) error {
	if !isValidPath(configPath) {
		return errors.New("invalid file path")
	}

	file, err := os.Open(configPath) //#nosec
	if err != nil {
		return err
	}

	decoder := yaml.NewDecoder(file)
	err = decoder.Decode(cfg)
	if err != nil {
		return err
	}

	return nil
}

func isValidPath(path string) bool {
	if path == "" {
		return false
	}

	// Check if the path exists
	_, err := os.Stat(path)
	if err != nil {
		if os.IsNotExist(err) {
			return false
		}
	}

	// Check if the path points to a file
	info, err := os.Stat(path)
	if err != nil {
		return false
	}
	if info.IsDir() {
		return false
	}

	return true
}
