// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"io"
	"os"
	"strings"
	"testing"
)

func Test_PrintBuildInfo(t *testing.T) {
	// Create a pipe to capture standard output
	readPipe, writePipe, _ := os.Pipe()
	defer readPipe.Close()

	// Redirect standard output to the write end of the pipe
	oldStdout := os.Stdout
	os.Stdout = writePipe

	PrintBuildInfo()

	// Restore standard output
	os.Stdout = oldStdout
	writePipe.Close()

	// Read the captured output from the read end of the pipe
	var outputBytes []byte
	outputBytes, _ = io.ReadAll(readPipe)

	if !strings.Contains(string(outputBytes), "Version:") {
		t.Errorf("Unexpected help message.\nGot:\n%s", string(outputBytes))
	}
}

func Test_PrintHelp(t *testing.T) {
	// Create a pipe to capture standard output
	readPipe, writePipe, _ := os.Pipe()
	defer readPipe.Close()

	// Redirect standard output to the write end of the pipe
	oldStdout := os.Stdout
	os.Stdout = writePipe

	PrintHelp()

	// Restore standard output
	os.Stdout = oldStdout
	writePipe.Close()

	// Read the captured output from the read end of the pipe
	var outputBytes []byte
	outputBytes, _ = io.ReadAll(readPipe)

	expectedHelp := `Possible arguments:
  --config		Changes default path from where to read the config file.
  --help		Print this help statement.
  --version		Print version and build info.`

	if strings.TrimSpace(string(outputBytes)) != strings.TrimSpace(expectedHelp) {
		t.Errorf("Unexpected help message.\nExpected:\n%s\nGot:\n%s", expectedHelp, string(outputBytes))
	}
}

func Test_ReadConfig(t *testing.T) {
	loadConfig := func(_ testing.TB, path string) error {
		t.Helper()

		return ReadConfig(&Cfg, path)
	}

	// Test case 1
	t.Run("test non existing config file", func(t *testing.T) {
		err := loadConfig(t, "config.yml")

		if err.Error() != "invalid file path" {
			t.Errorf("Unexpected error occurred: %v", err)
		}
	})

	// Test case 2
	t.Run("test unmarshal error", func(t *testing.T) {
		err := loadConfig(t, "../../config/pwch.service")

		if !strings.Contains(err.Error(), "yaml: unmarshal errors") {
			t.Errorf("Unexpected error occurred: %v", err)
		}
	})

	// Test case 3
	t.Run("test successful case", func(t *testing.T) {
		err := loadConfig(t, "../../config/config.yml")
		if err != nil {
			t.Errorf("Expected error to be nil, but got: %v", err)
		}

		expectedDomain := "example.com"
		if Cfg.Domain != expectedDomain {
			t.Errorf("Expected Domain: %s, Got: %s", expectedDomain, Cfg.Domain)
		}
	})
}

func Test_isValidPath(t *testing.T) {
	tests := []struct {
		name string
		path string
		want bool
	}{
		{
			name: "test empty path",
			path: "",
			want: false,
		},
		{
			name: "test non existing path",
			path: "/doesnotexist",
			want: false,
		},
		{
			name: "test if file",
			path: "/home",
			want: false,
		},
		{
			name: "test successful case",
			path: "../../config/config.yml",
			want: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := isValidPath(tt.path); got != tt.want {
				t.Errorf("isValidPath() = %v, want %v", got, tt.want)
			}
		})
	}
}
