// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"crypto/rand"
	"database/sql"
	"encoding/base64"
	"errors"
	"log"
	"net/mail"
	"net/smtp"
	"strings"
	"sync"
	"time"
)

const linebreak = "\r\n"

// used to fetch account attributes from database.
type mailUser struct {
	Enabled  bool
	Username string
	Domain   string
}

// OneTimeURLs is where valid one time URLs are stored
//
// key   = random token + username + domain
// value = time at creation of entry
//
// entries are deleted either after the password
// got changed or when the entry expires.
var OneTimeURLs = struct { //nolint:exhaustruct
	sync.RWMutex
	M map[string]time.Time
}{M: make(map[string]time.Time)}

func DeleteFromHashMap(key string) {
	OneTimeURLs.RLock()
	delete(OneTimeURLs.M, key)
	OneTimeURLs.RUnlock()
}

func addToHashMap(key string, value time.Time) {
	OneTimeURLs.Lock()
	OneTimeURLs.M[key] = value
	OneTimeURLs.Unlock()
}

func genRandomBytes(n int) ([]byte, error) {
	b := make([]byte, n)
	if _, err := rand.Read(b); err != nil {
		return nil, err
	}

	return b, nil
}

func genRandomString(n int) (string, error) {
	b, err := genRandomBytes(n)

	return base64.URLEncoding.EncodeToString(b), err
}

func isValidEmail(email string) bool {
	_, err := mail.ParseAddress(email)

	return err == nil
}

func sendOneTimeLink(username, domain string) {
	token, err := genRandomString(Cfg.OTL.TokenLength)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot generate random string")

		return
	}

	loginUser := Cfg.SMTP.LoginUser
	loginPassword := Cfg.SMTP.LoginPassword
	from := Cfg.SMTP.Sender
	to := []string{username + "@" + domain}
	host := Cfg.SMTP.Host
	port := Cfg.SMTP.Port

	accessString := "changePassword?token=" + token +
		"&username=" + username + "&domain=" + domain

	message := []byte("From: " + from + linebreak +
		"To: " + username + "@" + domain + linebreak +
		"Subject: Password change requested" + linebreak +
		linebreak +
		"Follow this link to change your password:" + linebreak +
		linebreak +
		"https://" + Cfg.Domain + Cfg.URLPrefix + "/" + accessString + linebreak +
		linebreak +
		"It's valid for 10 minutes." + linebreak +
		linebreak +
		"If you did not request a password change then just disregard this message." + linebreak)

	auth := smtp.PlainAuth("", loginUser, loginPassword, host)

	err = smtp.SendMail(host+":"+port, auth, loginUser, to, message)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: Sending OTL failed")

		return
	}

	addToHashMap(accessString, time.Now())
	log.Print("INFO: Sent OTL to " + username + "@" + domain)
}

func emailEnabled(email string) (bool, mailUser, error) {
	components := strings.Split(email, "@")
	username, domain := components[0], components[1]

	db, err := connectToDatabase(&Cfg)
	if err != nil {
		return false, mailUser{}, err
	}
	defer db.Close()

	stmt, err := db.Prepare("SELECT username, domain, enabled FROM accounts WHERE username = $1 AND domain = $2;")
	if err != nil {
		return false, mailUser{}, err
	}
	defer stmt.Close()

	var user mailUser

	err = stmt.QueryRow(username, domain).Scan(&user.Username, &user.Domain, &user.Enabled)
	if err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			log.Print("INFO: Unknown email address: " + username + "@" + domain)

			return false, user, nil
		}

		return false, user, err
	}

	if user.Enabled {
		log.Print("INFO: " + username + "@" + domain + " successfully validated")

		return true, user, nil
	}

	return false, user, nil
}
