// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"bytes"
	"crypto/rand"
	"errors"
	"log"
	"os"
	"regexp"
	"testing"
	"time"
)

// custom error reader.
type errorReader struct{}

func (r errorReader) Read(_ []byte) (int, error) {
	return 0, errors.New("custom error")
}

func Test_DeleteFromHashMap(t *testing.T) {
	m := OneTimeURLs.M
	key := "test_key"
	value := time.Now()

	// Add a value to the map
	m[key] = value

	// Call the function to delete the key from the map
	DeleteFromHashMap(key)

	// Check that the key has been removed
	if _, ok := m[key]; ok {
		t.Errorf("Expected key '%s' to be deleted, but it still exists in the map", key)
	}
}

func Test_addToHashMap(t *testing.T) {
	m := OneTimeURLs.M
	key := "test_key"
	value := time.Now()

	// Call the function to add a key-value pair to the map
	addToHashMap(key, value)

	// Check that the key-value pair exists in the map
	if val, ok := m[key]; !ok || val != value {
		t.Errorf("Expected key-value pair '%s:%v' to be added to the map, but it is not present or the value is incorrect", key, value)
	}
}

func TestGenRandomBytes(t *testing.T) {
	t.Run("unexpected length", func(t *testing.T) {
		length := 10
		randomBytes, err := genRandomBytes(length)
		if err != nil {
			t.Errorf("Error generating random bytes: %v", err)
		}
		if len(randomBytes) != length {
			t.Errorf("Unexpected length of random bytes.\nExpected: %d\nGot: %d", length, len(randomBytes))
		}
	})

	t.Run("unexpected error", func(t *testing.T) {
		// Replace the default rand.Reader with the errorReader.
		originalReader := rand.Reader
		rand.Reader = errorReader{}

		randomBytes, err := genRandomBytes(10)

		// Restore the original rand.Reader
		rand.Reader = originalReader

		if err == nil {
			t.Error("Expected error, but got nil")
		} else {
			expectedError := "custom error"
			if err.Error() != expectedError {
				t.Errorf("Unexpected error.\nExpected: %s\nGot: %s", expectedError, err.Error())
			}
		}

		if randomBytes != nil {
			t.Errorf("Expected nil byte slice, but got: %v", randomBytes)
		}
	})
}

func TestIsValidEmail(t *testing.T) {
	// Test case 1
	email := "test@example.com"
	result := isValidEmail(email)
	if result == false {
		t.Errorf("Got %s is invail but should be vaild", email)
	}

	// Test case 2
	email = "testexample.com"
	result = isValidEmail(email)
	if result == true {
		t.Errorf("Got %s is vail but should be invaild", email)
	}
}

func Test_sendOneTimeLink(t *testing.T) {
	Cfg.OTL.TokenLength = 64
	Cfg.SMTP.LoginUser = "noreply@localdomain"
	Cfg.SMTP.LoginPassword = "password"
	Cfg.SMTP.Sender = "PWCH <noreply@localdomain"
	Cfg.SMTP.Host = "localhost"
	Cfg.SMTP.Port = "587"
	Cfg.Domain = "localdomain"
	Cfg.URLPrefix = ""

	testSMTP := func(_ testing.TB, username, domain string) string {
		t.Helper()

		var buf bytes.Buffer
		log.SetOutput(&buf)

		sendOneTimeLink(username, domain)

		// Get the log output from the buffer
		output := buf.String()

		// Remove the date and timestamp portion from the log messages
		re := regexp.MustCompile(`\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} `)
		cleanedOutput := re.ReplaceAllString(output, "")

		log.SetOutput(os.Stdout)

		return cleanedOutput
	}

	// Test case 1
	t.Run("test successful delivery", func(t *testing.T) {
		got := testSMTP(t, "pwch1", "localdomain")
		want := "INFO: Sent OTL to pwch1@localdomain\n"

		if got != want {
			t.Errorf("\nGot:\n%s\nWant:\n%s", got, want)
		}
	})

	// Test case 2
	t.Run("test failed delivery", func(t *testing.T) {
		got := testSMTP(t, "test", "127.0.0.1")
		want := "501 5.1.3 Bad recipient address syntax\n" +
			"ERROR: Sending OTL failed\n"

		if got != want {
			t.Errorf("\nGot:\n%s\nWant:\n%s", got, want)
		}
	})
}

func Test_emailEnabled(t *testing.T) {
	Cfg.DB.Host = "/run/postgresql"
	Cfg.DB.DBName = "vmail"
	Cfg.DB.User = "vmail"
	Cfg.DB.Password = "password"

	testEmailEnabled := func(_ testing.TB, email string) (bool, string, error) {
		t.Helper()

		var buf bytes.Buffer
		log.SetOutput(&buf)

		enabled, _, err := emailEnabled(email)

		// Get the log output from the buffer
		output := buf.String()

		// Remove the date and timestamp portion from the log messages
		re := regexp.MustCompile(`\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} `)
		cleanedOutput := re.ReplaceAllString(output, "")

		log.SetOutput(os.Stdout)

		return enabled, cleanedOutput, err
	}

	// Test case 1: Valid email
	t.Run("Valid Email", func(t *testing.T) {
		enabled, gotOutput, err := testEmailEnabled(t, "pwch1@localdomain")
		wantOutput := "INFO: pwch1@localdomain successfully validated\n"

		if !enabled {
			t.Errorf("Expected email to be enabled, but got disabled")
		}

		if gotOutput != wantOutput {
			t.Errorf("\nGot:\n%s\nWant:\n%s", gotOutput, wantOutput)
		}

		if err != nil {
			t.Error(err)
		}
	})

	// Test case 2: Unknown email
	t.Run("Unknown Email", func(t *testing.T) {
		enabled, gotOutput, err := testEmailEnabled(t, "unknown@localdomain")
		wantOutput := "INFO: Unknown email address: unknown@localdomain\n"

		if enabled {
			t.Errorf("Expected email to be disabled, but got enabled")
		}

		if gotOutput != wantOutput {
			t.Errorf("\nGot:\n%s\nWant:\n%s", gotOutput, wantOutput)
		}

		if err != nil {
			t.Error(err)
		}
	})
}
