// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"bytes"
	"context"
	"database/sql"
	"encoding/hex"
	"errors"
	"fmt"
	"log"
	"os/exec"
	"unicode"

	"golang.org/x/crypto/bcrypt"
	"golang.org/x/crypto/sha3"
)

func hashPassword(password string) (string, error) {
	hash, err := bcrypt.GenerateFromPassword([]byte(password), Cfg.Bcrypt.Cost)

	return string(hash), err
}

// returns true if password and hash match, otherwise false.
func checkPasswordHash(password, hash string) bool {
	err := bcrypt.CompareHashAndPassword([]byte(hash), []byte(password))
	if err != nil {
		log.Print(err)

		return false
	}

	return true
}

func passwordMatches(username, domain, oldPass string) bool {
	db, err := connectToDatabase(&Cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	var hash string
	if err := db.QueryRow("SELECT password FROM accounts WHERE username = $1 AND domain = $2;",
		username, domain).Scan(&hash); err != nil {
		if err == sql.ErrNoRows {
			return false
		}
	}

	if checkPasswordHash(oldPass, hash) {
		log.Print("INFO: Successfully validated old password for " + username + "@" + domain)

		return true
	}
	log.Print("ERROR: Can't validate old password for " + username + "@" + domain)

	return false
}

func reencryptKeys(username, domain, email, oldPass, newPass string) error {
	db, err := connectToDatabase(&Cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	var mailCryptSalt string
	if queryErr := db.QueryRow("SELECT mail_crypt_salt FROM accounts WHERE username = $1 AND domain = $2;",
		username, domain).Scan(&mailCryptSalt); queryErr != nil {
		return queryErr
	}

	oldHash := sha3.Sum512([]byte(mailCryptSalt + oldPass))
	newHash := sha3.Sum512([]byte(mailCryptSalt + newPass))

	oldHashString := hex.EncodeToString(oldHash[:])
	newHashString := hex.EncodeToString(newHash[:])

	cmd := exec.Command("/usr/bin/pwch-doveadm-wrapper", "swap")

	var input bytes.Buffer
	input.WriteString(email + "\n" + oldHashString + "\n" + newHashString + "\n" + newHashString + "\n")

	cmd.Stdin = &input

	err = cmd.Run()
	if err == nil {
		log.Printf("INFO: Successfully swapped keys for %s", email)

		return nil
	}

	log.Printf("ERROR: Can't swap keys for %s", email)
	log.Print(err)

	return err
}

func terminateIMAPSessions(email string) error {
	cmd := exec.Command("/usr/bin/pwch-doveadm-wrapper", "kick", email)
	err := cmd.Run()

	if err == nil {
		log.Printf("INFO: Successfully terminated all sessions for %s", email)

		return nil
	}

	var exitErr *exec.ExitError

	if errors.As(err, &exitErr) && exitErr.ExitCode() == 68 {
		log.Printf("INFO: No active sessions to terminate for %s", email)

		return nil
	}

	log.Printf("ERROR: Can't terminate sessions for %s", email)
	log.Print(err)

	return err
}

func validatePasswordFields(newPass, confirmPass, oldPass string) error {
	if newPass != confirmPass {
		return errors.New("Passwords do not match") //nolint:stylecheck
	}

	if oldPass == newPass {
		return errors.New("You are trying to set the same password again") //nolint:stylecheck
	}

	return nil
}

func enforcePasswordPolicy(password string) (bool, string) {
	if len(password) < Cfg.PasswordPolicy.MinLength {
		return false, fmt.Sprintf("Please enter at least a %d character long password", Cfg.PasswordPolicy.MinLength)
	}

	if len(password) > Cfg.PasswordPolicy.MaxLength {
		return false, fmt.Sprintf("Please enter at max a %d character long password", Cfg.PasswordPolicy.MaxLength)
	}

	// initialize with opposite of config value
	// policies set to "false" will therefore init with "true" and will not change
	hasLower := !Cfg.PasswordPolicy.LowerCase
	hasUpper := !Cfg.PasswordPolicy.UpperCase
	hasNumber := !Cfg.PasswordPolicy.Digits
	hasSpecial := !Cfg.PasswordPolicy.SepcialChar

	for _, char := range password {
		switch {
		case unicode.IsNumber(char):
			hasNumber = true
		case unicode.IsLower(char):
			hasLower = true
		case unicode.IsUpper(char):
			hasUpper = true
		case unicode.IsPunct(char), unicode.IsSpace(char), unicode.IsSymbol(char):
			hasSpecial = true
		}
	}

	if hasLower && hasUpper && hasNumber && hasSpecial {
		return true, "Success"
	}

	errorMessage := "Undefined error"
	switch {
	case !hasLower:
		errorMessage = "Please enter at least one lower case character"
	case !hasUpper:
		errorMessage = "Please enter at least one upper case character"
	case !hasNumber:
		errorMessage = "Please enter at least one digit"
	case !hasSpecial:
		errorMessage = "Please enter at least one special character"
	}

	return false, errorMessage
}

// updates password in database, reencrypts user keys and terminates IMAP sessions.
func updatePassword(username, domain, newPass, oldPass string) error {
	if !passwordMatches(username, domain, oldPass) {
		return errors.New("Current Password does not match") //nolint:stylecheck
	}

	hash, err := hashPassword(newPass)
	if err != nil {
		log.Print(err)

		return err
	}

	db, err := connectToDatabase(&Cfg)
	if err != nil {
		log.Print(err)
		log.Print("ERROR: cannot connect to database")
	}
	defer db.Close()

	ctx := context.Background()
	tx, err := db.BeginTx(ctx, nil)
	if err != nil {
		log.Print("ERROR: can't begin transaction")

		return err
	}

	defer func() {
		if rollbackErr := tx.Rollback(); rollbackErr != nil {
			log.Print(rollbackErr)
			log.Print("ERROR: cannot rollback transaction")
		}
	}()

	_, err = tx.ExecContext(ctx, "UPDATE accounts SET password = $1 WHERE username = $2 AND domain = $3;",
		hash, username, domain)
	if err != nil {
		log.Print("ERROR: password update query failed")

		return err
	}

	email := username + "@" + domain
	if err = reencryptKeys(username, domain, email, oldPass, newPass); err != nil {
		return errors.New("internal error: password not changed")
	}

	err = tx.Commit()
	if err != nil {
		return err
	}

	log.Print("INFO: Password successfully changed for " + email)
	err = terminateIMAPSessions(email)

	return err
}
