// Copyright (C) 2023  Benedikt Zumtobel
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package pwch

import (
	"bytes"
	"fmt"
	"log"
	"os"
	"regexp"
	"strings"
	"testing"
)

func Test_enforcePasswordPolicy(t *testing.T) {
	Cfg.PasswordPolicy.MinLength = 12
	Cfg.PasswordPolicy.MaxLength = 24
	Cfg.PasswordPolicy.LowerCase = true
	Cfg.PasswordPolicy.UpperCase = true
	Cfg.PasswordPolicy.Digits = true
	Cfg.PasswordPolicy.SepcialChar = true

	testPassword := func(_ testing.TB, password, expectedMessage string) {
		t.Helper()

		valid, message := enforcePasswordPolicy(password)

		if valid {
			t.Errorf("Expected invalid password for input '%s', but got valid", password)
		}

		if message != expectedMessage {
			t.Errorf("Expected error message '%s', but got: %s", expectedMessage, message)
		}
	}

	// Test case 1: Valid password that meets all requirements
	password := "StrongPassword123!"
	valid, message := enforcePasswordPolicy(password)
	if !valid {
		t.Errorf("Expected valid password for input '%s', but got invalid", password)
	}
	if message != "Success" {
		t.Errorf("Expected success message for valid password, but got: %s", message)
	}

	// Test case 2
	t.Run("password of insufficient length", func(t *testing.T) {
		password = "short"
		expectedMessage := fmt.Sprintf("Please enter at least a %d character long password", Cfg.PasswordPolicy.MinLength)
		testPassword(t, password, expectedMessage)
	})

	// Test case 3
	t.Run("password exceeding maximum length", func(t *testing.T) {
		password = "thispasswordexceedsthemaximumallowedlength"
		expectedMessage := fmt.Sprintf("Please enter at max a %d character long password", Cfg.PasswordPolicy.MaxLength)
		testPassword(t, password, expectedMessage)
	})

	// Test case 4
	t.Run("password missing lower case character", func(t *testing.T) {
		password = "PASSWORD123!"
		expectedMessage := "Please enter at least one lower case character"
		testPassword(t, password, expectedMessage)
	})

	// Test case 5
	t.Run("password missing upper case character", func(t *testing.T) {
		password = "password123!"
		expectedMessage := "Please enter at least one upper case character"
		testPassword(t, password, expectedMessage)
	})

	// Test case 6
	t.Run("password missing digit", func(t *testing.T) {
		password = "PasswordWithoutDigit!"
		expectedMessage := "Please enter at least one digit"
		testPassword(t, password, expectedMessage)
	})

	// Test case 7
	t.Run("password missing special character", func(t *testing.T) {
		password = "PasswordNoSpecial123"
		expectedMessage := "Please enter at least one special character"
		testPassword(t, password, expectedMessage)
	})
}

func Test_validatePasswordFields(t *testing.T) {
	newPass := "newPassword"
	confirmPass := "newPassword"
	oldPass := "oldPassword" //nolint:goconst

	// Test case 1: Valid input
	err := validatePasswordFields(newPass, confirmPass, oldPass)
	if err != nil {
		t.Errorf("Expected no error for valid input, but got: %s", err.Error())
	}

	// Test case 2: Mismatched passwords
	confirmPass = "wrongPassword"
	err = validatePasswordFields(newPass, confirmPass, oldPass)
	if err == nil {
		t.Error("Expected error for mismatched passwords, but got no error")
	} else if err.Error() != "Passwords do not match" {
		t.Errorf("Expected error message 'Passwords do not match', but got: %s", err.Error())
	}

	// Test case 3: Same old and new passwords
	newPass = "oldPassword"
	confirmPass = "oldPassword"
	err = validatePasswordFields(newPass, confirmPass, oldPass)
	if err == nil {
		t.Error("Expected error for setting same password again, but got no error")
	} else if err.Error() != "You are trying to set the same password again" {
		t.Errorf("Expected error message 'You are trying to set the same password again', but got: %s", err.Error())
	}
}

func Test_reencryptKeys(t *testing.T) {
	Cfg.DB.Host = "/run/postgresql"
	Cfg.DB.DBName = "vmail"
	Cfg.DB.User = "vmail"
	Cfg.DB.Password = "password"
	Cfg.DB.SSLMode = "disable"

	testReencryption := func(_ testing.TB, username, domain, email, oldPassword, newPassword string) (string, error) {
		t.Helper()

		var buf bytes.Buffer
		log.SetOutput(&buf)

		err := reencryptKeys(username, domain, email, oldPassword, newPassword)

		// Get the log output from the buffer
		output := buf.String()

		// Remove the date and timestamp portion from the log messages
		re := regexp.MustCompile(`\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} `)
		cleanedOutput := re.ReplaceAllString(output, "")

		log.SetOutput(os.Stdout)

		return cleanedOutput, err
	}

	// Test case 1
	t.Run("successful reencryption", func(t *testing.T) {
		logMessage, err := testReencryption(t, "pwch2", "localdomain", "pwch2@localdomain", "password", "StrongPassword1234!")
		if err != nil {
			t.Errorf("want nil but got %v", err)
		}

		if !strings.Contains(logMessage, "Successfully") {
			t.Errorf("got unexpected log message: %s", logMessage)
		}
	})

	// Test case 2
	t.Run("failed reencryption of same user again", func(t *testing.T) {
		logMessage, err := testReencryption(t, "pwch2", "localdomain", "pwch2@localdomain", "password", "StrongPassword1234!")
		if err == nil {
			t.Errorf("want error but got nil")
		}

		if !strings.Contains(logMessage, "exit status 65") {
			t.Errorf("got unexpected log message: %s", logMessage)
		}
	})

	// Test case 3
	t.Run("failed reencryption of another user", func(t *testing.T) {
		logMessage, err := testReencryption(t, "pwch3", "localdomain", "pwch3@localdomain", "password123", "StrongPassword1234!")
		if err == nil {
			t.Errorf("want error but got nil")
		}

		if !strings.Contains(logMessage, "exit status 65") {
			t.Errorf("got unexpected log message: %s", logMessage)
		}
	})

	// Test case 4
	t.Run("non existing user", func(t *testing.T) {
		logMessage, err := testReencryption(t, "test", "localdomain", "test@localdomain", "password", "StrongPassword1234!")
		if err.Error() != "sql: no rows in result set" {
			t.Errorf("want error message 'sql: no rows in result set', but got: %s", err.Error())
		}

		if !strings.Contains(logMessage, "") {
			t.Errorf("got unexpected log message: %s", logMessage)
		}
	})

	// Test case 5
	t.Run("revert test case 1", func(t *testing.T) {
		logMessage, err := testReencryption(t, "pwch2", "localdomain", "pwch2@localdomain", "StrongPassword1234!", "password")
		if err != nil {
			t.Errorf("want nil but got %v", err)
		}

		if !strings.Contains(logMessage, "Successfully") {
			t.Errorf("got unexpected log message: %s", logMessage)
		}
	})
}
